# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding: utf-8
# pylint: disable=too-many-arguments, too-many-locals
"""Definition of various recurrent neural network cells."""

import bisect
import random
import numpy as np

from ..io import DataIter, DataBatch, DataDesc
from .. import ndarray

def encode_sentences(sentences, vocab=None, invalid_label=-1, invalid_key='\n',
                     start_label=0, unknown_token=None):
    """Encode sentences and (optionally) build a mapping
    from string tokens to integer indices. Unknown keys
    will be added to vocabulary.

    Parameters
    ----------
    sentences : list of list of str
        A list of sentences to encode. Each sentence
        should be a list of string tokens.
    vocab : None or dict of str -> int
        Optional input Vocabulary
    invalid_label : int, default -1
        Index for invalid token, like <end-of-sentence>
    invalid_key : str, default '\\n'
        Key for invalid token. Use '\\n' for end
        of sentence by default.
    start_label : int
        lowest index.
    unknown_token: str
        Symbol to represent unknown token.
        If not specified, unknown token will be skipped.

    Returns
    -------
    result : list of list of int
        encoded sentences
    vocab : dict of str -> int
        result vocabulary
    """
    idx = start_label
    if vocab is None:
        vocab = {invalid_key: invalid_label}
        new_vocab = True
    else:
        new_vocab = False
    res = []
    for sent in sentences:
        coded = []
        for word in sent:
            if word not in vocab:
                assert (new_vocab or unknown_token), "Unknown token %s"%word
                if idx == invalid_label:
                    idx += 1
                if unknown_token:
                    word = unknown_token
                vocab[word] = idx
                idx += 1
            coded.append(vocab[word])
        res.append(coded)

    return res, vocab

class BucketSentenceIter(DataIter):
    """Simple bucketing iterator for language model.
    The label at each sequence step is the following token
    in the sequence.

    Parameters
    ----------
    sentences : list of list of int
        Encoded sentences.
    batch_size : int
        Batch size of the data.
    invalid_label : int, optional
        Key for invalid label, e.g. <end-of-sentence>. The default is -1.
    dtype : str, optional
        Data type of the encoding. The default data type is 'float32'.
    buckets : list of int, optional
        Size of the data buckets. Automatically generated if None.
    data_name : str, optional
        Name of the data. The default name is 'data'.
    label_name : str, optional
        Name of the label. The default name is 'softmax_label'.
    layout : str, optional
        Format of data and label. 'NT' means (batch_size, length)
        and 'TN' means (length, batch_size).
    """
    def __init__(self, sentences, batch_size, buckets=None, invalid_label=-1,
                 data_name='data', label_name='softmax_label', dtype='float32',
                 layout='NT'):
        super(BucketSentenceIter, self).__init__()
        if not buckets:
            buckets = [i for i, j in enumerate(np.bincount([len(s) for s in sentences]))
                       if j >= batch_size]
        buckets.sort()

        ndiscard = 0
        self.data = [[] for _ in buckets]
        valid_buckets = {}
        for item in range(len(buckets)):
            valid_buckets[item] = 0

        for i, sent in enumerate(sentences):
            buck = bisect.bisect_left(buckets, len(sent))
            valid_buckets[buck] = 1
            if buck == len(buckets):
                ndiscard += 1
                continue
            buff = np.full((buckets[buck],), invalid_label, dtype=dtype)
            buff[:len(sent)] = sent
            self.data[buck].append(buff)
        buckets = [j for i, j in enumerate(buckets) if valid_buckets[i] == 1]
        self.data = [np.asarray(i, dtype=dtype) for i in self.data if i]

        print("WARNING: discarded %d sentences longer than the largest bucket."%ndiscard)

        self.batch_size = batch_size
        self.buckets = buckets
        self.data_name = data_name
        self.label_name = label_name
        self.dtype = dtype
        self.invalid_label = invalid_label
        self.nddata = []
        self.ndlabel = []
        self.major_axis = layout.find('N')
        self.layout = layout
        self.default_bucket_key = max(buckets)

        if self.major_axis == 0:
            self.provide_data = [DataDesc(
                name=self.data_name, shape=(batch_size, self.default_bucket_key),
                layout=self.layout)]
            self.provide_label = [DataDesc(
                name=self.label_name, shape=(batch_size, self.default_bucket_key),
                layout=self.layout)]
        elif self.major_axis == 1:
            self.provide_data = [DataDesc(
                name=self.data_name, shape=(self.default_bucket_key, batch_size),
                layout=self.layout)]
            self.provide_label = [DataDesc(
                name=self.label_name, shape=(self.default_bucket_key, batch_size),
                layout=self.layout)]
        else:
            raise ValueError("Invalid layout %s: Must by NT (batch major) or TN (time major)")

        self.idx = []
        for i, buck in enumerate(self.data):
            self.idx.extend([(i, j) for j in range(0, len(buck) - batch_size + 1, batch_size)])
        self.curr_idx = 0

        self.reset()

    def reset(self):
        """Resets the iterator to the beginning of the data."""
        self.curr_idx = 0
        random.shuffle(self.idx)
        for buck in self.data:
            np.random.shuffle(buck)

        self.nddata = []
        self.ndlabel = []
        for buck in self.data:
            label = np.empty_like(buck)
            label[:, :-1] = buck[:, 1:]
            label[:, -1] = self.invalid_label
            self.nddata.append(ndarray.array(buck, dtype=self.dtype))
            self.ndlabel.append(ndarray.array(label, dtype=self.dtype))

    def next(self):
        """Returns the next batch of data."""
        if self.curr_idx == len(self.idx):
            raise StopIteration
        i, j = self.idx[self.curr_idx]
        self.curr_idx += 1

        if self.major_axis == 1:
            data = self.nddata[i][j:j+self.batch_size].T
            label = self.ndlabel[i][j:j+self.batch_size].T
        else:
            data = self.nddata[i][j:j+self.batch_size]
            label = self.ndlabel[i][j:j+self.batch_size]

        return DataBatch([data], [label], pad=0,
                         bucket_key=self.buckets[i],
                         provide_data=[DataDesc(
                             name=self.data_name, shape=data.shape,
                             layout=self.layout)],
                         provide_label=[DataDesc(
                             name=self.label_name, shape=label.shape,
                             layout=self.layout)])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Utils for registering NumPy array function protocol for mxnet.numpy ops."""

import functools
import numpy as _np
from . import numpy as mx_np  # pylint: disable=reimported
from .numpy.multiarray import _NUMPY_ARRAY_FUNCTION_DICT, _NUMPY_ARRAY_UFUNC_DICT


def _find_duplicate(strs):
    str_set = set()
    for s in strs:
        if s in str_set:
            return s
        else:
            str_set.add(s)
    return None


def _implements(numpy_function):
    """Register an __array_function__ implementation for MyArray objects."""
    def decorator(func):
        _NUMPY_ARRAY_FUNCTION_DICT[numpy_function] = func
        return func
    return decorator


def with_array_function_protocol(func):
    """A decorator for functions that expect array function protocol.
    The decorated function only runs when NumPy version >= 1.17."""
    from distutils.version import LooseVersion
    cur_np_ver = LooseVersion(_np.__version__)
    np_1_17_ver = LooseVersion('1.17')

    @functools.wraps(func)
    def _run_with_array_func_proto(*args, **kwargs):
        if cur_np_ver >= np_1_17_ver:
            try:
                func(*args, **kwargs)
            except Exception as e:
                raise RuntimeError('Running function {} with NumPy array function protocol failed'
                                   ' with exception {}'
                                   .format(func.__name__, str(e)))

    return _run_with_array_func_proto


def with_array_ufunc_protocol(func):
    """A decorator for functions that expect array ufunc protocol.
    The decorated function only runs when NumPy version >= 1.15."""
    from distutils.version import LooseVersion
    cur_np_ver = LooseVersion(_np.__version__)
    np_1_15_ver = LooseVersion('1.15')

    @functools.wraps(func)
    def _run_with_array_ufunc_proto(*args, **kwargs):
        if cur_np_ver >= np_1_15_ver:
            try:
                func(*args, **kwargs)
            except Exception as e:
                raise RuntimeError('Running function {} with NumPy array ufunc protocol failed'
                                   ' with exception {}'
                                   .format(func.__name__, str(e)))

    return _run_with_array_ufunc_proto


_NUMPY_ARRAY_FUNCTION_LIST = [
    'all',
    'any',
    'sometrue',
    'argmin',
    'argmax',
    'around',
    'round',
    'round_',
    'argsort',
    'sort',
    'append',
    'broadcast_arrays',
    'broadcast_to',
    'clip',
    'concatenate',
    'copy',
    'cumsum',
    'diag',
    'diagonal',
    'diagflat',
    'dot',
    'expand_dims',
    'fix',
    'flip',
    'flipud',
    'fliplr',
    'inner',
    'insert',
    'max',
    'amax',
    'mean',
    'min',
    'amin',
    'nonzero',
    'ones_like',
    'atleast_1d',
    'atleast_2d',
    'atleast_3d',
    'prod',
    'product',
    'ravel',
    'repeat',
    'reshape',
    'roll',
    'split',
    'array_split',
    'hsplit',
    'vsplit',
    'dsplit',
    'squeeze',
    'stack',
    'std',
    'sum',
    'swapaxes',
    'take',
    'tensordot',
    'tile',
    'transpose',
    'unique',
    'unravel_index',
    'diag_indices_from',
    'delete',
    'var',
    'vdot',
    'vstack',
    'column_stack',
    'hstack',
    'dstack',
    'zeros_like',
    'linalg.norm',
    'linalg.cholesky',
    'linalg.inv',
    'linalg.solve',
    'linalg.tensorinv',
    'linalg.tensorsolve',
    'linalg.pinv',
    'linalg.eigvals',
    'linalg.eig',
    'linalg.eigvalsh',
    'linalg.eigh',
    'shape',
    'trace',
    'tril',
    'meshgrid',
    'outer',
    'einsum',
    'polyval',
    'shares_memory',
    'may_share_memory',
    'quantile',
    'percentile',
    'diff',
    'ediff1d',
    'resize',
    'where',
    'full_like',
    'bincount',
    'empty_like',
    'nan_to_num',
    'isnan',
    'isfinite',
    'isposinf',
    'isneginf',
    'isinf',
    'pad',
]


@with_array_function_protocol
def _register_array_function():
    """Register __array_function__ protocol for mxnet.numpy operators so that
    ``mxnet.numpy.ndarray`` can be fed into the official NumPy operators and
    dispatched to MXNet implementation.

    Notes
    -----
    According the __array_function__ protocol (see the following reference),
    there are three kinds of operators that cannot be dispatched using this
    protocol:
    1. Universal functions, which already have their own protocol in the official
    NumPy package.
    2. Array creation functions.
    3. Dispatch for methods of any kind, e.g., methods on np.random.RandomState objects.

    References
    ----------
    https://numpy.org/neps/nep-0018-array-function-protocol.html
    """
    dup = _find_duplicate(_NUMPY_ARRAY_FUNCTION_LIST)
    if dup is not None:
        raise ValueError('Duplicate operator name {} in _NUMPY_ARRAY_FUNCTION_LIST'.format(dup))
    for op_name in _NUMPY_ARRAY_FUNCTION_LIST:
        strs = op_name.split('.')
        if len(strs) == 1:
            mx_np_op = getattr(mx_np, op_name)
            onp_op = getattr(_np, op_name)
            setattr(mx_np, op_name, _implements(onp_op)(mx_np_op))
        elif len(strs) == 2:
            mx_np_submodule = getattr(mx_np, strs[0])
            mx_np_op = getattr(mx_np_submodule, strs[1])
            onp_submodule = getattr(_np, strs[0])
            onp_op = getattr(onp_submodule, strs[1])
            setattr(mx_np_submodule, strs[1], _implements(onp_op)(mx_np_op))
        else:
            raise ValueError('Does not support registering __array_function__ protocol '
                             'for operator {}'.format(op_name))


# https://docs.scipy.org/doc/numpy/reference/ufuncs.html#available-ufuncs
_NUMPY_ARRAY_UFUNC_LIST = [
    'abs',
    'fabs',
    'add',
    'arctan2',
    'copysign',
    'degrees',
    'hypot',
    'lcm',
    # 'ldexp',
    'subtract',
    'multiply',
    'true_divide',
    'negative',
    'power',
    'mod',
    'matmul',
    'absolute',
    'rint',
    'sign',
    'exp',
    'log',
    'log2',
    'log10',
    'expm1',
    'sqrt',
    'square',
    'cbrt',
    'reciprocal',
    'invert',
    'bitwise_not',
    'remainder',
    'sin',
    'cos',
    'tan',
    'sinh',
    'cosh',
    'tanh',
    'arcsin',
    'arccos',
    'arctan',
    'arcsinh',
    'arccosh',
    'arctanh',
    'maximum',
    'minimum',
    'ceil',
    'trunc',
    'floor',
    'bitwise_and',
    'bitwise_xor',
    'bitwise_or',
    'logical_not',
    'equal',
    'not_equal',
    'less',
    'less_equal',
    'greater',
    'greater_equal',
]


@with_array_ufunc_protocol
def _register_array_ufunc():
    """Register NumPy array ufunc protocol.

    References
    ----------
    https://numpy.org/neps/nep-0013-ufunc-overrides.html
    """
    dup = _find_duplicate(_NUMPY_ARRAY_UFUNC_LIST)
    if dup is not None:
        raise ValueError('Duplicate operator name {} in _NUMPY_ARRAY_UFUNC_LIST'.format(dup))
    for op_name in _NUMPY_ARRAY_UFUNC_LIST:
        try:
            mx_np_op = getattr(mx_np, op_name)
            _NUMPY_ARRAY_UFUNC_DICT[op_name] = mx_np_op
        except AttributeError:
            raise AttributeError('mxnet.numpy does not have operator named {}'.format(op_name))


_register_array_function()
_register_array_ufunc()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Numpy basic functions."""

from .stride_tricks import broadcast_arrays

__all__ = ['meshgrid']


def meshgrid(*xi, **kwargs):
    """
    Return coordinate matrices from coordinate vectors.

    Make N-D coordinate arrays for vectorized evaluations of
    N-D scalar/vector fields over N-D grids, given
    one-dimensional coordinate arrays x1, x2,..., xn.

    Parameters
    ----------
    x1, x2,..., xn : ndarrays
        1-D arrays representing the coordinates of a grid.
    indexing : {'xy', 'ij'}, optional
        Cartesian ('xy', default) or matrix ('ij') indexing of output.
        See Notes for more details.

    sparse : bool, optional
        If True a sparse grid is returned in order to conserve memory.
        Default is False. Please note that `sparse=True` is currently
        not supported.

    copy : bool, optional
        If False, a view into the original arrays are returned in order to
        conserve memory.  Default is True. Please note that `copy=False`
        is currently not supported.

    Returns
    -------
    X1, X2,..., XN : ndarray
        For vectors `x1`, `x2`,..., 'xn' with lengths ``Ni=len(xi)`` ,
        return ``(N1, N2, N3,...Nn)`` shaped arrays if indexing='ij'
        or ``(N2, N1, N3,...Nn)`` shaped arrays if indexing='xy'
        with the elements of `xi` repeated to fill the matrix along
        the first dimension for `x1`, the second for `x2` and so on.

    Notes
    -----
    This function supports both indexing conventions through the indexing
    keyword argument.  Giving the string 'ij' returns a meshgrid with
    matrix indexing, while 'xy' returns a meshgrid with Cartesian indexing.
    In the 2-D case with inputs of length M and N, the outputs are of shape
    (N, M) for 'xy' indexing and (M, N) for 'ij' indexing.  In the 3-D case
    with inputs of length M, N and P, outputs are of shape (N, M, P) for
    'xy' indexing and (M, N, P) for 'ij' indexing.  The difference is
    illustrated by the following code snippet::

        xv, yv = np.meshgrid(x, y, sparse=False, indexing='ij')
        for i in range(nx):
            for j in range(ny):
                # treat xv[i,j], yv[i,j]

        xv, yv = np.meshgrid(x, y, sparse=False, indexing='xy')
        for i in range(nx):
            for j in range(ny):
                # treat xv[j,i], yv[j,i]

    In the 1-D and 0-D case, the indexing and sparse keywords have no effect.
    """
    ndim = len(xi)

    copy_ = kwargs.pop('copy', True)
    if not copy_:
        raise NotImplementedError('copy=False is not implemented')
    sparse = kwargs.pop('sparse', False)
    if sparse:
        raise NotImplementedError('sparse=False is not implemented')
    indexing = kwargs.pop('indexing', 'xy')

    if kwargs:
        raise TypeError("meshgrid() got an unexpected keyword argument '%s'"
                        % (list(kwargs)[0],))

    if indexing not in ['xy', 'ij']:
        raise ValueError(
            "Valid values for `indexing` are 'xy' and 'ij'.")

    s0 = (1,) * ndim
    output = [x.reshape(s0[:i] + (-1,) + s0[i + 1:])
              for i, x in enumerate(xi)]

    if indexing == 'xy' and ndim > 1:
        # switch first and second axis
        output[0] = output[0].reshape(1, -1, *s0[2:])
        output[1] = output[1].reshape(-1, 1, *s0[2:])

    if not sparse:
        # Return the full N-D matrix (not only the 1-D vector)
        output = broadcast_arrays(*output)

    return output



# coding: utf-8# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# File content is auto-generated. Do not modify.
# pylint: skip-file
from ._internal import NDArrayBase
from ..base import _Null

def adjust_lighting(data=None, alpha=_Null, out=None, name=None, **kwargs):
    r"""Adjust the lighting level of the input. Follow the AlexNet style.

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L254

    Parameters
    ----------
    data : NDArray
        The input.
    alpha : tuple of <float>, required
        The lighting alphas for the R, G, B channels.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def crop(data=None, x=_Null, y=_Null, width=_Null, height=_Null, out=None, name=None, **kwargs):
    r"""Crop an image NDArray of shape (H x W x C) or (N x H x W x C) 
    to the given size.
    Example:
        .. code-block:: python
            image = mx.nd.random.uniform(0, 255, (4, 2, 3)).astype(dtype=np.uint8)
            mx.nd.image.crop(image, 1, 1, 2, 2)
                [[[144  34   4]
                  [ 82 157  38]]

                 [[156 111 230]
                  [177  25  15]]]
                <NDArray 2x2x3 @cpu(0)>
            image = mx.nd.random.uniform(0, 255, (2, 4, 2, 3)).astype(dtype=np.uint8)
            mx.nd.image.crop(image, 1, 1, 2, 2)            
                [[[[ 35 198  50]
                   [242  94 168]]

                  [[223 119 129]
                   [249  14 154]]]


                  [[[137 215 106]
                    [ 79 174 133]]

                   [[116 142 109]
                    [ 35 239  50]]]]
                <NDArray 2x2x2x3 @cpu(0)>


    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/crop.cc:L65

    Parameters
    ----------
    data : NDArray
        The input.
    x : int, required
        Left boundary of the cropping area.
    y : int, required
        Top boundary of the cropping area.
    width : int, required
        Width of the cropping area.
    height : int, required
        Height of the cropping area.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def flip_left_right(data=None, out=None, name=None, **kwargs):
    r"""

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L195

    Parameters
    ----------
    data : NDArray
        The input.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def flip_top_bottom(data=None, out=None, name=None, **kwargs):
    r"""

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L205

    Parameters
    ----------
    data : NDArray
        The input.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def normalize(data=None, mean=_Null, std=_Null, out=None, name=None, **kwargs):
    r"""Normalize an tensor of shape (C x H x W) or (N x C x H x W) with mean and
        standard deviation.

        Given mean `(m1, ..., mn)` and std `(s\ :sub:`1`\ , ..., s\ :sub:`n`)` for `n` channels,
        this transform normalizes each channel of the input tensor with:

    .. math::

            output[i] = (input[i] - m\ :sub:`i`\ ) / s\ :sub:`i`

        If mean or std is scalar, the same value will be applied to all channels.

        Default value for mean is 0.0 and stand deviation is 1.0.

    Example:

        .. code-block:: python
            image = mx.nd.random.uniform(0, 1, (3, 4, 2))
            normalize(image, mean=(0, 1, 2), std=(3, 2, 1))
                [[[ 0.18293785  0.19761486]
                  [ 0.23839645  0.28142193]
                  [ 0.20092112  0.28598186]
                  [ 0.18162774  0.28241724]]
                 [[-0.2881726  -0.18821815]
                  [-0.17705294 -0.30780914]
                  [-0.2812064  -0.3512327 ]
                  [-0.05411351 -0.4716435 ]]
                 [[-1.0363373  -1.7273437 ]
                  [-1.6165586  -1.5223348 ]
                  [-1.208275   -1.1878313 ]
                  [-1.4711051  -1.5200229 ]]]
                <NDArray 3x4x2 @cpu(0)>

            image = mx.nd.random.uniform(0, 1, (2, 3, 4, 2))
            normalize(image, mean=(0, 1, 2), std=(3, 2, 1))
                [[[[ 0.18934818  0.13092826]
                   [ 0.3085322   0.27869293]
                   [ 0.02367868  0.11246539]
                   [ 0.0290431   0.2160573 ]]
                  [[-0.4898908  -0.31587923]
                   [-0.08369008 -0.02142242]
                   [-0.11092162 -0.42982462]
                   [-0.06499392 -0.06495637]]
                  [[-1.0213816  -1.526392  ]
                   [-1.2008414  -1.1990893 ]
                   [-1.5385206  -1.4795225 ]
                   [-1.2194707  -1.3211205 ]]]
                 [[[ 0.03942481  0.24021089]
                   [ 0.21330701  0.1940066 ]
                   [ 0.04778443  0.17912441]
                   [ 0.31488964  0.25287187]]
                  [[-0.23907584 -0.4470462 ]
                   [-0.29266903 -0.2631998 ]
                   [-0.3677222  -0.40683383]
                   [-0.11288315 -0.13154092]]
                  [[-1.5438497  -1.7834496 ]
                   [-1.431566   -1.8647819 ]
                   [-1.9812102  -1.675859  ]
                   [-1.3823645  -1.8503251 ]]]]
                <NDArray 2x3x4x2 @cpu(0)>


    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L167

    Parameters
    ----------
    data : NDArray
        Input ndarray
    mean : tuple of <float>, optional, default=[0,0,0,0]
        Sequence of means for each channel. Default value is 0.
    std : tuple of <float>, optional, default=[1,1,1,1]
        Sequence of standard deviations for each channel. Default value is 1.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def random_brightness(data=None, min_factor=_Null, max_factor=_Null, out=None, name=None, **kwargs):
    r"""

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L215

    Parameters
    ----------
    data : NDArray
        The input.
    min_factor : float, required
        Minimum factor.
    max_factor : float, required
        Maximum factor.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def random_color_jitter(data=None, brightness=_Null, contrast=_Null, saturation=_Null, hue=_Null, out=None, name=None, **kwargs):
    r"""

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L246

    Parameters
    ----------
    data : NDArray
        The input.
    brightness : float, required
        How much to jitter brightness.
    contrast : float, required
        How much to jitter contrast.
    saturation : float, required
        How much to jitter saturation.
    hue : float, required
        How much to jitter hue.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def random_contrast(data=None, min_factor=_Null, max_factor=_Null, out=None, name=None, **kwargs):
    r"""

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L222

    Parameters
    ----------
    data : NDArray
        The input.
    min_factor : float, required
        Minimum factor.
    max_factor : float, required
        Maximum factor.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def random_flip_left_right(data=None, out=None, name=None, **kwargs):
    r"""

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L200

    Parameters
    ----------
    data : NDArray
        The input.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def random_flip_top_bottom(data=None, out=None, name=None, **kwargs):
    r"""

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L210

    Parameters
    ----------
    data : NDArray
        The input.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def random_hue(data=None, min_factor=_Null, max_factor=_Null, out=None, name=None, **kwargs):
    r"""

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L238

    Parameters
    ----------
    data : NDArray
        The input.
    min_factor : float, required
        Minimum factor.
    max_factor : float, required
        Maximum factor.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def random_lighting(data=None, alpha_std=_Null, out=None, name=None, **kwargs):
    r"""Randomly add PCA noise. Follow the AlexNet style.

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L262

    Parameters
    ----------
    data : NDArray
        The input.
    alpha_std : float, optional, default=0.0500000007
        Level of the lighting noise.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def random_saturation(data=None, min_factor=_Null, max_factor=_Null, out=None, name=None, **kwargs):
    r"""

    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L230

    Parameters
    ----------
    data : NDArray
        The input.
    min_factor : float, required
        Minimum factor.
    max_factor : float, required
        Maximum factor.

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def resize(data=None, size=_Null, keep_ratio=_Null, interp=_Null, out=None, name=None, **kwargs):
    r"""Resize an image NDArray of shape (H x W x C) or (N x H x W x C) 
    to the given size
    Example:
        .. code-block:: python
            image = mx.nd.random.uniform(0, 255, (4, 2, 3)).astype(dtype=np.uint8)
            mx.nd.image.resize(image, (3, 3))
                [[[124 111 197]
                  [158  80 155]
                  [193  50 112]]

                 [[110 100 113]
                  [134 165 148]
                  [157 231 182]]

                 [[202 176 134]
                  [174 191 149]
                  [147 207 164]]]
                <NDArray 3x3x3 @cpu(0)>
            image = mx.nd.random.uniform(0, 255, (2, 4, 2, 3)).astype(dtype=np.uint8)
            mx.nd.image.resize(image, (2, 2))            
                [[[[ 59 133  80]
                   [187 114 153]]

                  [[ 38 142  39]
                   [207 131 124]]]


                  [[[117 125 136]
                   [191 166 150]]

                  [[129  63 113]
                   [182 109  48]]]]
                <NDArray 2x2x2x3 @cpu(0)>


    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/resize.cc:L70

    Parameters
    ----------
    data : NDArray
        The input.
    size : Shape(tuple), optional, default=[]
        Size of new image. Could be (width, height) or (size)
    keep_ratio : boolean, optional, default=0
        Whether to resize the short edge or both edges to `size`, if size is give as an integer.
    interp : int, optional, default='1'
        Interpolation method for resizing. By default uses bilinear interpolationOptions are INTER_NEAREST - a nearest-neighbor interpolationINTER_LINEAR - a bilinear interpolationINTER_AREA - resampling using pixel area relationINTER_CUBIC - a bicubic interpolation over 4x4 pixel neighborhoodINTER_LANCZOS4 - a Lanczos interpolation over 8x8 pixel neighborhoodNote that the GPU version only support bilinear interpolation(1)

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

def to_tensor(data=None, out=None, name=None, **kwargs):
    r"""Converts an image NDArray of shape (H x W x C) or (N x H x W x C) 
    with values in the range [0, 255] to a tensor NDArray of shape (C x H x W) or (N x C x H x W)
    with values in the range [0, 1]

    Example:
        .. code-block:: python
            image = mx.nd.random.uniform(0, 255, (4, 2, 3)).astype(dtype=np.uint8)
            to_tensor(image)
                [[[ 0.85490197  0.72156864]
                  [ 0.09019608  0.74117649]
                  [ 0.61960787  0.92941177]
                  [ 0.96470588  0.1882353 ]]
                 [[ 0.6156863   0.73725492]
                  [ 0.46666667  0.98039216]
                  [ 0.44705883  0.45490196]
                  [ 0.01960784  0.8509804 ]]
                 [[ 0.39607844  0.03137255]
                  [ 0.72156864  0.52941179]
                  [ 0.16470589  0.7647059 ]
                  [ 0.05490196  0.70588237]]]
                 <NDArray 3x4x2 @cpu(0)>

            image = mx.nd.random.uniform(0, 255, (2, 4, 2, 3)).astype(dtype=np.uint8)
            to_tensor(image)
                [[[[0.11764706 0.5803922 ]
                   [0.9411765  0.10588235]
                   [0.2627451  0.73333335]
                   [0.5647059  0.32156864]]
                  [[0.7176471  0.14117648]
                   [0.75686276 0.4117647 ]
                   [0.18431373 0.45490196]
                   [0.13333334 0.6156863 ]]
                  [[0.6392157  0.5372549 ]
                   [0.52156866 0.47058824]
                   [0.77254903 0.21568628]
                   [0.01568628 0.14901961]]]
                 [[[0.6117647  0.38431373]
                   [0.6784314  0.6117647 ]
                   [0.69411767 0.96862745]
                   [0.67058825 0.35686275]]
                  [[0.21960784 0.9411765 ]
                   [0.44705883 0.43529412]
                   [0.09803922 0.6666667 ]
                   [0.16862746 0.1254902 ]]
                  [[0.6156863  0.9019608 ]
                   [0.35686275 0.9019608 ]
                   [0.05882353 0.6509804 ]
                   [0.20784314 0.7490196 ]]]]
                <NDArray 2x3x4x2 @cpu(0)>


    Defined in /ci_py311_second_pass/mxnet-suite_1679346683348/work/src/operator/image/image_random.cc:L92

    Parameters
    ----------
    data : NDArray
        Input ndarray

    out : NDArray, optional
        The output NDArray to hold the result.

    Returns
    -------
    out : NDArray or list of NDArrays
        The output of this function.
    """
    return (0,)

__all__ = ['adjust_lighting', 'crop', 'flip_left_right', 'flip_top_bottom', 'normalize', 'random_brightness', 'random_color_jitter', 'random_contrast', 'random_flip_left_right', 'random_flip_top_bottom', 'random_hue', 'random_lighting', 'random_saturation', 'resize', 'to_tensor']


# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding: utf-8
# pylint: disable= arguments-differ
"""Custom convolutional neural network layers in model_zoo."""

__all__ = ['DeformableConvolution', 'ModulatedDeformableConvolution']

from .... import symbol
from ...block import HybridBlock
from ....base import numeric_types
from ...nn import Activation

class DeformableConvolution(HybridBlock):
    """2-D Deformable Convolution v_1 (Dai, 2017).
    Normal Convolution uses sampling points in a regular grid, while the sampling
    points of Deformablem Convolution can be offset. The offset is learned with a
    separate convolution layer during the training. Both the convolution layer for
    generating the output features and the offsets are included in this gluon layer.

    Parameters
    ----------
    channels : int,
        The dimensionality of the output space
        i.e. the number of output channels in the convolution.
    kernel_size : int or tuple/list of 2 ints, (Default value = (1,1))
        Specifies the dimensions of the convolution window.
    strides : int or tuple/list of 2 ints, (Default value = (1,1))
        Specifies the strides of the convolution.
    padding : int or tuple/list of 2 ints, (Default value = (0,0))
        If padding is non-zero, then the input is implicitly zero-padded
        on both sides for padding number of points.
    dilation : int or tuple/list of 2 ints, (Default value = (1,1))
        Specifies the dilation rate to use for dilated convolution.
    groups : int, (Default value = 1)
        Controls the connections between inputs and outputs.
        At groups=1, all inputs are convolved to all outputs.
        At groups=2, the operation becomes equivalent to having two convolution
        layers side by side, each seeing half the input channels, and producing
        half the output channels, and both subsequently concatenated.
    num_deformable_group : int, (Default value = 1)
        Number of deformable group partitions.
    layout : str, (Default value = NCHW)
        Dimension ordering of data and weight. Can be 'NCW', 'NWC', 'NCHW',
        'NHWC', 'NCDHW', 'NDHWC', etc. 'N', 'C', 'H', 'W', 'D' stands for
        batch, channel, height, width and depth dimensions respectively.
        Convolution is performed over 'D', 'H', and 'W' dimensions.
    use_bias : bool, (Default value = True)
        Whether the layer for generating the output features uses a bias vector.
    in_channels : int, (Default value = 0)
        The number of input channels to this layer. If not specified,
        initialization will be deferred to the first time `forward` is called
        and input channels will be inferred from the shape of input data.
    activation : str, (Default value = None)
        Activation function to use. See :func:`~mxnet.ndarray.Activation`.
        If you don't specify anything, no activation is applied
        (ie. "linear" activation: `a(x) = x`).
    weight_initializer : str or `Initializer`, (Default value = None)
        Initializer for the `weight` weights matrix for the convolution layer
        for generating the output features.
    bias_initializer : str or `Initializer`, (Default value = zeros)
        Initializer for the bias vector for the convolution layer
        for generating the output features.
    offset_weight_initializer : str or `Initializer`, (Default value = zeros)
        Initializer for the `weight` weights matrix for the convolution layer
        for generating the offset.
    offset_bias_initializer : str or `Initializer`, (Default value = zeros),
        Initializer for the bias vector for the convolution layer
        for generating the offset.
    offset_use_bias: bool, (Default value = True)
        Whether the layer for generating the offset uses a bias vector.

    Inputs:
        - **data**: 4D input tensor with shape
          `(batch_size, in_channels, height, width)` when `layout` is `NCHW`.
          For other layouts shape is permuted accordingly.

    Outputs:
        - **out**: 4D output tensor with shape
          `(batch_size, channels, out_height, out_width)` when `layout` is `NCHW`.
          out_height and out_width are calculated as::

              out_height = floor((height+2*padding[0]-dilation[0]*(kernel_size[0]-1)-1)/stride[0])+1
              out_width = floor((width+2*padding[1]-dilation[1]*(kernel_size[1]-1)-1)/stride[1])+1
    """

    def __init__(self, channels, kernel_size=(1, 1), strides=(1, 1), padding=(0, 0), dilation=(1, 1), groups=1,
                 num_deformable_group=1, layout='NCHW', use_bias=True, in_channels=0, activation=None,
                 weight_initializer=None, bias_initializer='zeros',
                 offset_weight_initializer='zeros', offset_bias_initializer='zeros', offset_use_bias=True,
                 op_name='DeformableConvolution', adj=None, prefix=None, params=None):
        super(DeformableConvolution, self).__init__(prefix=prefix, params=params)
        with self.name_scope():
            self._channels = channels
            self._in_channels = in_channels

            assert layout in ('NCHW', 'NHWC'), "Only supports 'NCHW' and 'NHWC' layout for now"
            if isinstance(kernel_size, numeric_types):
                kernel_size = (kernel_size,) * 2
            if isinstance(strides, numeric_types):
                strides = (strides,) * len(kernel_size)
            if isinstance(padding, numeric_types):
                padding = (padding,) * len(kernel_size)
            if isinstance(dilation, numeric_types):
                dilation = (dilation,) * len(kernel_size)
            self._op_name = op_name

            offset_channels = 2 * kernel_size[0] * kernel_size[1] * num_deformable_group
            self._kwargs_offset = {
                'kernel': kernel_size, 'stride': strides, 'dilate': dilation,
                'pad': padding, 'num_filter': offset_channels, 'num_group': groups,
                'no_bias': not offset_use_bias, 'layout': layout}

            self._kwargs_deformable_conv = {
                'kernel': kernel_size, 'stride': strides, 'dilate': dilation,
                'pad': padding, 'num_filter': channels, 'num_group': groups,
                'num_deformable_group': num_deformable_group,
                'no_bias': not use_bias, 'layout': layout}

            if adj:
                self._kwargs_offset['adj'] = adj
                self._kwargs_deformable_conv['adj'] = adj

            dshape = [0] * (len(kernel_size) + 2)
            dshape[layout.find('N')] = 1
            dshape[layout.find('C')] = in_channels

            op = getattr(symbol, 'Convolution')
            offset = op(symbol.var('data', shape=dshape), **self._kwargs_offset)

            offsetshapes = offset.infer_shape_partial()[0]

            self.offset_weight = self.params.get('offset_weight', shape=offsetshapes[1],
                                                 init=offset_weight_initializer,
                                                 allow_deferred_init=True)

            if offset_use_bias:
                self.offset_bias = self.params.get('offset_bias', shape=offsetshapes[2],
                                                   init=offset_bias_initializer,
                                                   allow_deferred_init=True)
            else:
                self.offset_bias = None

            deformable_conv_weight_shape = [0] * (len(kernel_size) + 2)
            deformable_conv_weight_shape[0] = channels
            deformable_conv_weight_shape[2] = kernel_size[0]
            deformable_conv_weight_shape[3] = kernel_size[1]

            self.deformable_conv_weight = self.params.get('deformable_conv_weight',
                                                          shape=deformable_conv_weight_shape,
                                                          init=weight_initializer,
                                                          allow_deferred_init=True)

            if use_bias:
                self.deformable_conv_bias = self.params.get('deformable_conv_bias', shape=(channels,),
                                                            init=bias_initializer,
                                                            allow_deferred_init=True)
            else:
                self.deformable_conv_bias = None

            if activation:
                self.act = Activation(activation, prefix=activation + '_')
            else:
                self.act = None

    def hybrid_forward(self, F, x, offset_weight, deformable_conv_weight, offset_bias=None, deformable_conv_bias=None):
        if offset_bias is None:
            offset = F.Convolution(x, offset_weight, cudnn_off=True, **self._kwargs_offset)
        else:
            offset = F.Convolution(x, offset_weight, offset_bias, cudnn_off=True, **self._kwargs_offset)

        if deformable_conv_bias is None:
            act = F.contrib.DeformableConvolution(data=x, offset=offset, weight=deformable_conv_weight,
                                                  name='fwd', **self._kwargs_deformable_conv)
        else:
            act = F.contrib.DeformableConvolution(data=x, offset=offset, weight=deformable_conv_weight,
                                                  bias=deformable_conv_bias, name='fwd',
                                                  **self._kwargs_deformable_conv)

        if self.act:
            act = self.act(act)
        return act

    def _alias(self):
        return 'deformable_conv'

    def __repr__(self):
        s = '{name}({mapping}, kernel_size={kernel}, stride={stride}'
        len_kernel_size = len(self._kwargs_deformable_conv['kernel'])
        if self._kwargs_deformable_conv['pad'] != (0,) * len_kernel_size:
            s += ', padding={pad}'
        if self._kwargs_deformable_conv['dilate'] != (1,) * len_kernel_size:
            s += ', dilation={dilate}'
        if hasattr(self, 'out_pad') and self.out_pad != (0,) * len_kernel_size:
            s += ', output_padding={out_pad}'.format(out_pad=self.out_pad)
        if self._kwargs_deformable_conv['num_group'] != 1:
            s += ', groups={num_group}'
        if self.deformable_conv_bias is None:
            s += ', bias=False'
        if self.act:
            s += ', {}'.format(self.act)
        s += ')'
        shape = self.deformable_conv_weight.shape
        return s.format(name=self.__class__.__name__,
                        mapping='{0} -> {1}'.format(shape[1] if shape[1] else None, shape[0]),
                        **self._kwargs_deformable_conv)


class ModulatedDeformableConvolution(HybridBlock):
    """2-D Deformable Convolution v2 (Dai, 2018).

    The modulated deformable convolution operation is described in https://arxiv.org/abs/1811.11168

    Parameters
    ----------
    channels : int,
        The dimensionality of the output space
        i.e. the number of output channels in the convolution.
    kernel_size : int or tuple/list of 2 ints, (Default value = (1,1))
        Specifies the dimensions of the convolution window.
    strides : int or tuple/list of 2 ints, (Default value = (1,1))
        Specifies the strides of the convolution.
    padding : int or tuple/list of 2 ints, (Default value = (0,0))
        If padding is non-zero, then the input is implicitly zero-padded
        on both sides for padding number of points.
    dilation : int or tuple/list of 2 ints, (Default value = (1,1))
        Specifies the dilation rate to use for dilated convolution.
    groups : int, (Default value = 1)
        Controls the connections between inputs and outputs.
        At groups=1, all inputs are convolved to all outputs.
        At groups=2, the operation becomes equivalent to having two convolution
        layers side by side, each seeing half the input channels, and producing
        half the output channels, and both subsequently concatenated.
    num_deformable_group : int, (Default value = 1)
        Number of deformable group partitions.
    layout : str, (Default value = NCHW)
        Dimension ordering of data and weight. Can be 'NCW', 'NWC', 'NCHW',
        'NHWC', 'NCDHW', 'NDHWC', etc. 'N', 'C', 'H', 'W', 'D' stands for
        batch, channel, height, width and depth dimensions respectively.
        Convolution is performed over 'D', 'H', and 'W' dimensions.
    use_bias : bool, (Default value = True)
        Whether the layer for generating the output features uses a bias vector.
    in_channels : int, (Default value = 0)
        The number of input channels to this layer. If not specified,
        initialization will be deferred to the first time `forward` is called
        and input channels will be inferred from the shape of input data.
    activation : str, (Default value = None)
        Activation function to use. See :func:`~mxnet.ndarray.Activation`.
        If you don't specify anything, no activation is applied
        (ie. "linear" activation: `a(x) = x`).
    weight_initializer : str or `Initializer`, (Default value = None)
        Initializer for the `weight` weights matrix for the convolution layer
        for generating the output features.
    bias_initializer : str or `Initializer`, (Default value = zeros)
        Initializer for the bias vector for the convolution layer
        for generating the output features.
    offset_weight_initializer : str or `Initializer`, (Default value = zeros)
        Initializer for the `weight` weights matrix for the convolution layer
        for generating the offset.
    offset_bias_initializer : str or `Initializer`, (Default value = zeros),
        Initializer for the bias vector for the convolution layer
        for generating the offset.
    offset_use_bias: bool, (Default value = True)
        Whether the layer for generating the offset uses a bias vector.

    Inputs:
        - **data**: 4D input tensor with shape
          `(batch_size, in_channels, height, width)` when `layout` is `NCHW`.
          For other layouts shape is permuted accordingly.

    Outputs:
        - **out**: 4D output tensor with shape
          `(batch_size, channels, out_height, out_width)` when `layout` is `NCHW`.
          out_height and out_width are calculated as::

              out_height = floor((height+2*padding[0]-dilation[0]*(kernel_size[0]-1)-1)/stride[0])+1
              out_width = floor((width+2*padding[1]-dilation[1]*(kernel_size[1]-1)-1)/stride[1])+1
    """

    def __init__(self, channels, kernel_size=(1, 1), strides=(1, 1), padding=(0, 0), dilation=(1, 1), groups=1,
                 num_deformable_group=1, layout='NCHW', use_bias=True, in_channels=0, activation=None,
                 weight_initializer=None, bias_initializer='zeros',
                 offset_weight_initializer='zeros', offset_bias_initializer='zeros', offset_use_bias=True,
                 op_name='ModulatedDeformableConvolution', adj=None, prefix=None, params=None):
        super(ModulatedDeformableConvolution, self).__init__(prefix=prefix, params=params)
        with self.name_scope():
            self._channels = channels
            self._in_channels = in_channels

            assert layout in ('NCHW', 'NHWC'), "Only supports 'NCHW' and 'NHWC' layout for now"
            if isinstance(kernel_size, numeric_types):
                kernel_size = (kernel_size,) * 2
            if isinstance(strides, numeric_types):
                strides = (strides,) * len(kernel_size)
            if isinstance(padding, numeric_types):
                padding = (padding,) * len(kernel_size)
            if isinstance(dilation, numeric_types):
                dilation = (dilation,) * len(kernel_size)
            self._op_name = op_name

            offset_channels = num_deformable_group * 3 * kernel_size[0] * kernel_size[1]
            self.offset_split_index = num_deformable_group * 2 * kernel_size[0] * kernel_size[1]
            self._kwargs_offset = {
                'kernel': kernel_size, 'stride': strides, 'dilate': dilation,
                'pad': padding, 'num_filter': offset_channels, 'num_group': groups,
                'no_bias': not offset_use_bias, 'layout': layout}

            self._kwargs_deformable_conv = {
                'kernel': kernel_size, 'stride': strides, 'dilate': dilation,
                'pad': padding, 'num_filter': channels, 'num_group': groups,
                'num_deformable_group': num_deformable_group,
                'no_bias': not use_bias, 'layout': layout}

            if adj:
                self._kwargs_offset['adj'] = adj
                self._kwargs_deformable_conv['adj'] = adj

            deformable_conv_weight_shape = [0] * (len(kernel_size) + 2)
            deformable_conv_weight_shape[0] = channels
            deformable_conv_weight_shape[2] = kernel_size[0]
            deformable_conv_weight_shape[3] = kernel_size[1]

            self.deformable_conv_weight = self.params.get('deformable_conv_weight',
                                                          shape=deformable_conv_weight_shape,
                                                          init=weight_initializer,
                                                          allow_deferred_init=True)

            if use_bias:
                self.deformable_conv_bias = self.params.get('deformable_conv_bias', shape=(channels,),
                                                            init=bias_initializer,
                                                            allow_deferred_init=True)
            else:
                self.deformable_conv_bias = None

            dshape = [0] * (len(kernel_size) + 2)
            dshape[layout.find('N')] = 1
            dshape[layout.find('C')] = in_channels

            op = getattr(symbol, 'Convolution')
            offset = op(symbol.var('data', shape=dshape), **self._kwargs_offset)

            offsetshapes = offset.infer_shape_partial()[0]

            self.offset_weight = self.params.get('offset_weight', shape=offsetshapes[1],
                                                 init=offset_weight_initializer,
                                                 allow_deferred_init=True)

            if offset_use_bias:
                self.offset_bias = self.params.get('offset_bias', shape=offsetshapes[2],
                                                   init=offset_bias_initializer,
                                                   allow_deferred_init=True)
            else:
                self.offset_bias = None

            if activation:
                self.act = Activation(activation, prefix=activation + '_')
            else:
                self.act = None

    def hybrid_forward(self, F, x, offset_weight, deformable_conv_weight, offset_bias=None, deformable_conv_bias=None):
        if offset_bias is None:
            offset = F.Convolution(x, offset_weight, cudnn_off=True, **self._kwargs_offset)
        else:
            offset = F.Convolution(x, offset_weight, offset_bias, cudnn_off=True, **self._kwargs_offset)

        offset_t = F.slice_axis(offset, axis=1, begin=0, end=self.offset_split_index)
        mask = F.slice_axis(offset, axis=1, begin=self.offset_split_index, end=None)
        mask = F.sigmoid(mask) * 2

        if deformable_conv_bias is None:
            act = F.contrib.ModulatedDeformableConvolution(data=x, offset=offset_t, mask=mask,
                                                           weight=deformable_conv_weight,
                                                           name='fwd', **self._kwargs_deformable_conv)
        else:
            act = F.contrib.ModulatedDeformableConvolution(data=x, offset=offset_t, mask=mask,
                                                           weight=deformable_conv_weight,
                                                           bias=deformable_conv_bias, name='fwd',
                                                           **self._kwargs_deformable_conv)

        if self.act:
            act = self.act(act)
        return act

    def _alias(self):
        return 'modulated_deformable_conv'



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding: utf-8
# pylint: disable=no-member, protected-access, unused-import, no-name-in-module
# pylint: disable=wildcard-import, unused-wildcard-import
"""Random number interface of MXNet."""

import ctypes
from .base import _LIB, check_call, integer_types
from .ndarray.random import *
from .context import Context


def seed(seed_state, ctx="all"):
    """Seeds the random number generators in MXNet.

    This affects the behavior of modules in MXNet that uses random number generators,
    like the dropout operator and `NDArray`'s random sampling operators.

    Parameters
    ----------
    seed_state : int
        The random number seed.

    ctx : Context
        The device context of the generator. The default is "all" which means seeding random
        number generators of all devices.

    Notes
    -----
    Random number generators in MXNet are device specific.
    `mx.random.seed(seed_state)` sets the state of each generator using `seed_state` and the
    device id. Therefore, random numbers generated from different devices can be different
    even if they are seeded using the same seed.

    To produce identical random number sequences independent of the device id,
    set optional `ctx` argument. This produces the same sequence of random numbers independent
    of the device id, but the sequence can be different on different kind of devices as MXNet's
    random number generators for CPU and GPU use different algorithms.

    Example
    -------
    >>> print(mx.nd.random.normal(shape=(2,2)).asnumpy())
    [[ 1.36481571 -0.62203991]
     [-1.4962182  -0.08511394]]
    >>> print(mx.nd.random.normal(shape=(2,2)).asnumpy())
    [[ 1.09544981 -0.20014545]
     [-0.20808885  0.2527658 ]]
    # Same results on the same device with the same seed
    >>> mx.random.seed(128)
    >>> print(mx.nd.random.normal(shape=(2,2)).asnumpy())
    [[ 0.47400656 -0.75213492]
     [ 0.20251541  0.95352972]]
    >>> mx.random.seed(128)
    >>> print(mx.nd.random.normal(shape=(2,2)).asnumpy())
    [[ 0.47400656 -0.75213492]
     [ 0.20251541  0.95352972]]
    # Different results on gpu(0) and gpu(1) with the same seed
    >>> mx.random.seed(128)
    >>> print(mx.nd.random.normal(shape=(2,2), ctx=mx.gpu(0)).asnumpy())
    [[ 2.5020072 -1.6884501]
     [-0.7931333 -1.4218881]]
    >>> mx.random.seed(128)
    >>> print(mx.nd.random.normal(shape=(2,2), ctx=mx.gpu(1)).asnumpy())
    [[ 0.24336822 -1.664805  ]
     [-1.0223296   1.253198  ]]
    # Seeding with `ctx` argument produces identical results on gpu(0) and gpu(1)
    >>> mx.random.seed(128, ctx=mx.gpu(0))
    >>> print(mx.nd.random.normal(shape=(2,2), ctx=mx.gpu(0)).asnumpy())
    [[ 2.5020072 -1.6884501]
     [-0.7931333 -1.4218881]]
    >>> mx.random.seed(128, ctx=mx.gpu(1))
    >>> print(mx.nd.random.normal(shape=(2,2), ctx=mx.gpu(1)).asnumpy())
    [[ 2.5020072 -1.6884501]
     [-0.7931333 -1.4218881]]
    """
    if not isinstance(seed_state, integer_types):
        raise ValueError('seed_state must be int')
    seed_state = ctypes.c_int(int(seed_state))
    if ctx == "all":
        check_call(_LIB.MXRandomSeed(seed_state))
    else:
        ctx = Context(ctx)
        check_call(_LIB.MXRandomSeedContext(seed_state, ctx.device_typeid, ctx.device_id))
